# SnapMatcher.py
# Copyright 2005, 2006, 2007 Matt Heinzen

# This file is part of SnapMatcher 0.1.
#
# SnapMatcher is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# SnapMatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with File Duplicate Finder; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import Image
from scipy import *
import os, sys, string, re
from math import *

class ImageRecord:
	# fileSize: image filesize in bytes
	# coefs2x2: list of 2x2 grid signature coefficients (4 floats)
	# coefs4x4: list of 4x4 grid signature coefficients (24 floats)
	pass

class ImageListItem:
	# name: full file pathname
	# coefs2x2: list of 2x2 grid signature coefficients (4 floats)
	# coefs4x4: list of 4x4 grid signature coefficients (24 floats)
	pass


def absError(x, y):
	return x - y

# returns the average intensity for the specified subimage
def scanSubImage(pixels, width, height, minX, maxX, minY, maxY):
 	intensitySum = 0.0
 	for y in range(minY, maxY + 1):
 		rowIntensity = sum(pixels[y * width + minX:y * width + maxX + 1])
 		intensitySum += rowIntensity
 	return intensitySum / ((maxX - minX + 1) * (maxY - minY + 1))


# calulates a signature for the given image and grid resolution
# TODO: just scale down to exact size and eliminate need for scanSubImage (use
#   ANTIALIAS for maximum quality downsampling)
def calcImageSig(img, steps):
	# scale image down for speed
	w, h = img.size
	img = img.resize((w / 4, h / 4))

	# get new image size and pixel grid
	w, h = img.size
	pixels = list(img.getdata())

	grid = []
	for i in range(0, steps):
		rowsums = []
		minY = int((h * i) / steps)
		maxY = int((h * (i + 1)) / steps) - 1
		for j in range(0, steps):
			minX = int((w * j) / steps)
			maxX = int((w * (j + 1)) / steps) - 1
			rowsums.append(scanSubImage(pixels, w, h, minX, maxX, minY, maxY))
		grid.append(rowsums)

	# DIFFERENCE METHOD (does better than brightness when using hash table)
	coefs = []
	for row in range(0, steps):
		for col in range(0, steps - 1):
			coefs.append(absError(grid[row][col], grid[row][col + 1]))
	for row in range(0, steps - 1):
		for col in range(0, steps):
			coefs.append(absError(grid[row][col], grid[row + 1][col]))

	# BRIGHTNESS METHOD
	#coefs = []
	#for row in range(0, steps):
	#	for col in range(0, steps):
	#		coefs.append(grid[row][col])

	# TODO: test without this step
	# NOTE: using this step pushes individual coefficients outside of then
	#   (-256, 256) range, causing more spread which in a few cases will causes
	#   misses
	# ex. similar images with very different coefficients
	#   ALLEY02: 0.862226277914 29.3454936832 -27.3454936832 1.13777372209
	#   ALL9: -294.085419735 -151.918507609 153.918507609 296.085419735
	# correlations are identical with either method, so I should examine which
	#   has better spread properties (i.e. spreads different images while
	#   keeping similar images close)
	#avgError = sum(coefs) / len(coefs)
	#for i in range(0, len(coefs)): coefs[i] /= avgError

	return coefs


# adds DB entries for images in a directory
# used by os.path.walk
# DEPRECATED: now use os.walk instead of os.path.walk
def addImagesToDB(data, dirName, filenames):
	# data[0] is image DB
	# data[1] is extension regular expression
	# data[2] is mode (0 = quiet, 1 = terminal, 2 = progress)
	# data[3] is a callback func taking current dir as argument (mode 2 only)

	# TODO: deal with situation where file is already in DB but is out of date
	#   (file size in DB does not match current size)

	# don't walk hidden directories (directory names starting with '.')
	if os.path.split(dirName)[1][0] == '.': return

	# handle progress output
	if data[2] == 1: print "scanning " + dirName
	elif data[2] == 2:
		pass

	# add files to DB
	for name in filenames:
		if re.search(data[1], string.lower(name)) != None:
			path = os.path.join(dirName, name)
			if not os.path.isdir(path) and not data[0].has_key(path):
				try:
					img = Image.open(path)
					if img.getbands() != ('L'):
						img = img.convert('L')
					imgData = ImageRecord()
					imgData.fileSize = os.path.getsize(path)
					imgData.coefs2x2 = calcImageSig(img, 2)
					imgData.coefs4x4 = calcImageSig(img, 4)
					data[0][path] = imgData

					# DEBUG
					#print name + ": " + str(calcSigIndex(imgData.coefs2x2))
				except IOError:
					print "Unable to open image " + path


# creates a dictionary of image files
# dictionary maps file paths to image attributes (size and signatures)
def createImageDB(dirList, fileExts = [], mode = 1):
	# build a regular expression to determine if a filename matches any of the
	# provided file extensions, or any file if no extensions are provided
	# the pattern will apply to lowercase filenames only
	if len(fileExts) == 0:
		extRE = re.compile("")
	else:
		pattern = "\.("
		for ext in fileExts:
			pattern = pattern + ext + "|"
		pattern = pattern[0:-1] + ")$"
		extRE = re.compile(pattern)

	# build a regex to detect hidden directories
	pattern = os.sep + "\."
	hiddenRE = re.compile(pattern)

	imageDB = {}

	for directory in dirList:
		for root, dirs, files in os.walk(directory):
			if re.search(hiddenRE, root) == None:
				# handle progress output
				if mode == 1: print "scanning " + root \
					+ " (" + str(len(files)) + " files)"
				elif mode == 2:
					pass

				for name in files:
					if re.search(extRE, string.lower(name)) != None:
						path = os.path.join(root, name)
						if not imageDB.has_key(path):
							try:
								img = Image.open(path)
								if img.getbands() != ('L'):
									img = img.convert('L')
								imgData = ImageRecord()
								imgData.fileSize = os.path.getsize(path)
								imgData.coefs2x2 = calcImageSig(img, 2)
								imgData.coefs4x4 = calcImageSig(img, 4)
								imageDB[path] = imgData
							except IOError:
								print "Unable to open image " + path

	return imageDB


# writes the image database to a text file
def writeImageDBFile(imageDB, outFile):
	for key, val in imageDB.iteritems():
		elemStr = key + "*" + str(val.fileSize) + " "
		elemStr += string.join(map(str, val.coefs2x2)) + " "
		elemStr += string.join(map(str, val.coefs4x4)) + "\n"
		outFile.write(elemStr)


# reads the image database from a text file into a dictionary
def readImageDBFile(inFile):
	imageDB = {}
	for line in inFile.readlines():
		imgData = ImageRecord()
		name, rest = string.split(line, '*')
		words = string.split(rest)
		imgData.fileSize = int(words[0])
		imgData.coefs2x2 = map(float, words[1:5])
		imgData.coefs4x4 = map(float, words[5:])
		imageDB[name] = imgData
	return imageDB


# reads the image DB into a list (bypassing dictionary representation)
def readImageDBFileIntoList(inFile):
	imageList = []
	for line in inFile.readlines():
		imgItem = ImageListItem()
		imgItem.name, rest = string.split(line, '*')
		words = string.split(rest)
		#imgData.fileSize = int(words[0])
		imgItem.coefs2x2 = map(float, words[1:5])
		imgItem.coefs4x4 = map(float, words[5:])
		#imageDB[name] = imgData
		imageList.append(imgItem)
	return imageList


# converts the image database dictionary into a list (for faster iteration)
# DEPRECATED: DB now loaded directly into list, may be used again in future
def imageDB2List(imageDB):
	imageList = []
	for name, val in imageDB.iteritems():
		imgItem = ImageListItem()
		imgItem.name = name
		imgItem.coefs2x2 = val.coefs2x2
		imgItem.coefs4x4 = val.coefs4x4
		imageList.append(imgItem)
	return imageList


# TODO: make number of steps a parameter
HASH_STEPS = 1 # use powers of two here: 1 for all pairs testing up to 1024 for maximum speed
HASH_RANGE = 512.0 / HASH_STEPS

def calcSigIndex(coefs2x2):
	vals = map(lambda x: int((x + 256.0) / HASH_RANGE), coefs2x2)
	return (vals[0], vals[1], vals[2], vals[3])

def hashImages(imageList):
	imageHash = {}
	for i, img in enumerate(imageList):
		index = calcSigIndex(img.coefs2x2)
		if not imageHash.has_key(index):
			imageHash[index] = []
		imageHash[index].append(i)
	return imageHash


# generates a list of similar images from candidates in the image list based on
#   4x4 grid signature correlation
def genSimilarityList(imageList, minCorrTol = 0.9, mode = 1):
	minCorr = -1000.0
	maxCorr = 1000.0
	corrSum = 0.0
	totalComp = 0

	numScans = len(imageList) - 1
	numPairs = ((numScans + 1) * numScans) / 2

	simList = []

	for i, img1 in enumerate(imageList[:-1]):
		if mode == 1:
			print "starting scan " + str(i + 1) + " of " + str(numScans) \
				+ " (" + str(len(imageList[i + 1:])) + " images)"
		elif mode == 2:
			pass

		for img2 in imageList[i + 1:]:
			corr = stats.pearsonr(img1.coefs4x4, img2.coefs4x4)[0]
			if corr < minCorr: minCorr = corr
			if corr > maxCorr: maxCorr = corr
			corrSum += corr
			totalComp += 1
			if corr > minCorrTol: simList.append((corr, img1.name, img2.name))

		completePct = str(totalComp * 100.0 / numPairs)

		if mode == 1:
			print str(len(simList)) + " total image pairs found above " \
				"threshold (" + str(totalComp) + " pairs examined, " \
				+ completePct + "% complete)"
		elif mode == 2:
			pass

	return simList


def addKeys(a, b):
	return (a[0] + b[0], a[1] + b[1], a[2] + b[2], a[3] + b[3])

# generates a list of similar images from candidates in the image list based on
#   4x4 grid signature correlation
# to prevent O(n^2) time each image only searches among images in the local
#   cluster in space defined by 4 2x2 grid coefficients
def findSimilarImages(imageList, imageHash, minCorrTol = 0.9):
	simList = []

	totalComps = 0

	# TEST
	#minx = miny = minz = minw = 1000
	#maxx = maxy = maxz = maxw = -1000

	r = (-1, 0, 1)
	kernel = [(x, y, z, w) for x in r for y in r for z in r for w in r]

	for i, img in enumerate(imageList):
		if i % 100 == 0:

			# TODO: use a mode for output like DB creation
			print str(float(i) / len(imageList) * 100) + "% complete"

			#if mode == 1:
			#	print str(float(i) / len(imageList) * 100) + "% complete"
			#elif mode == 2:
			#	pass

		index = calcSigIndex(img.coefs2x2)

		# TEST
		#if index[0] < minx:
		#	minx = index[0]
		#	print img.coefs2x2
		#if index[1] < miny:
		#	miny = index[1]
		#	print img.coefs2x2
		#if index[2] < minz:
		#	minz = index[2]
		#	print img.coefs2x2
		#if index[3] < minw:
		#	minw = index[3]
		#	print img.coefs2x2
		#if index[0] > maxx:
		#	maxx = index[0]
		#	print img.coefs2x2
		#if index[1] > maxy:
		#	maxy = index[1]
		#	print img.coefs2x2
		#if index[2] > maxz:
		#	maxz = index[2]
		#	print img.coefs2x2
		#if index[3] > maxw:
		#	maxw = index[3]
		#	print img.coefs2x2
		#	print img.name

		for offset in kernel:
			key = addKeys(index, offset)
			if imageHash.has_key(key):
				imgCluster = imageHash[key]
				for j in imgCluster:
					if i < j:
						img2 = imageList[j]
						corr = stats.pearsonr(img.coefs4x4, img2.coefs4x4)[0]
						if corr > minCorrTol:
							simList.append((corr, img.name, img2.name))

						totalComps += 1

# 		for x in range(index[0] - 1, index[0] + 2):
# 			if x >= 0 and x < HASH_STEPS:
# 				for y in range(index[1] - 1, index[1] + 2):
# 					if y >= 0 and y < HASH_STEPS:
# 						for z in range(index[2] - 1, index[2] + 2):
# 							if z >= 0 and z < HASH_STEPS:
# 								for w in range(index[3] - 1, index[3] + 2):
# 									if w >= 0 and w < HASH_STEPS:
# 										key = (x, y, z, w)
# 										if imageHash.has_key(key):
# 											imgCluster = imageHash[key]
# 											for j in imgCluster:
# 												if i < j:
# 													img2 = imageList[j]
# 													corr = stats.pearsonr(
# 														img.coefs4x4,
# 														img2.coefs4x4)[0]
# 													if corr > minCorrTol:
# 														simList.append((corr,
# 															img.name,
# 															img2.name))
#
# 													totalComps += 1

	maxPairs = len(imageList) * (len(imageList) - 1) / 2
	print len(imageList), "images"
	print maxPairs, "max possible pairs"
	print totalComps, "comparisons made using hash table"
	print len(simList), "pairs found above threshold"
	print float(len(simList)) / totalComps, "efficiency"

	#print minx
	#print miny
	#print minz
	#print minw
	#print maxx
	#print maxy
	#print maxz
	#print maxw

	return simList


# write a list of similar image pairs to a text file
def writeSimilarityListFile(simList, outFile):
	for elem in simList:
		outFile.write(str(elem[0]) + "\n" + elem[1] + "\n" + elem[2] + "\n\n")


# returns a list of file extensions parsed from the given string
# extensions are separated, converted to lowercase, and stripped of all
#   separator and wildcard characters
def fileExtsFromString(fileExtStr = ""):
	exts = re.split("\W+", string.lower(fileExtStr))
	if len(exts) == 1 and len(exts[0]) == 0:
		exts = []
	return exts


if __name__ == "__main__":
	defExts = "jpg jpeg jpe gif tif tiff png bmp"
	#defHashSplits = 256
	defMinCorrTol = 0.95

	numArgs = len(sys.argv)

	# print USAGE and exit if invalid or incorrect number of arguments supplied
	if numArgs < 2 or (sys.argv[1] != "--builddb"
		and sys.argv[1] != "--finddups") or (sys.argv[1] == "--builddb"
		and numArgs < 4) or (sys.argv[1] == "--finddups" and numArgs < 3):
		print "USAGES:"
		print "    SnapMatcher.py --builddb <ROOTDIR> <GROUPNAME> [EXT 1] " \
			"[EXT 2] ..."
		print "    SnapMatcher.py --finddups <GROUPNAME> [TOLERANCE]"
		print "if no extensions specified for --builddb the following " \
			"default extensions will be used:", defExts
		print "if no minimum correlation tolerance specified for --finddups " \
			"then", defMinCorrTol, "will be used"
 		sys.exit()

	if sys.argv[1] == "--builddb":
		# convert root search directory to absolute path
		rootDir = os.path.abspath(sys.argv[2])
		groupName = sys.argv[3]
		imageDBFilename = "imagedb." + groupName

		# extract the valid file extensions to check for
		if numArgs > 4:
			fileExts = fileExtsFromString(string.join(sys.argv[4:], " "))
		else: fileExts = string.split(defExts)

		print "Building image database..."
		print "DIR=" + rootDir
		print "EXTS=" + str(fileExts)

		try:
			imageDBFile = open(imageDBFilename, "w")
			imageDB = createImageDB([rootDir], fileExts)
			writeImageDBFile(imageDB, imageDBFile)
			imageDBFile.close()
		except NameError:
			print "programmer screwed up!"
		except:
			print "ERROR: unable to write image database file '" \
				+ imageDBFilename + "'"
			sys.exit()

		print str(len(imageDB.keys())) + " images analyzed"

	elif sys.argv[1] == "--finddups":
		groupName = sys.argv[2]
		imageDBFilename = "imagedb." + groupName
		if numArgs > 3: minCorrTol = float(sys.argv[3])
		else: minCorrTol = defMinCorrTol

		try:
			imageDBFile = open(imageDBFilename, "r")
			imageList = readImageDBFileIntoList(imageDBFile)
			imageDBFile.close()
		except NameError:
			print "programmer screwed up!"
		except:
			print "ERROR: could not read image database file '" \
				+ imageDBFilename + "'"
			sys.exit()

		imageHash = hashImages(imageList)

		resultsFilename = "results." + groupName

		try:
			simFile = open(resultsFilename, "w")

			simList = findSimilarImages(imageList, imageHash, minCorrTol)

			simList.sort()
			simList.reverse()
			writeSimilarityListFile(simList, simFile)
			simFile.close()
		except NameError:
			print "programmer screwed up!"
		except:
			print "ERROR: unable to write file '" + resultsFilename + "'"
			sys.exit()

